/****************************************************************************
*
*  dcomp.cmd
*
*  Compare directories and synchronize content
*
*  Release: 0.13
*  Create:  01.09.1999
*  Update:  31.07.2003
*  (C)      Marcel Mller 1999-2003
*
*  Known problems:
*  - /Q does not work with Classic REXX. Use -Q or "/Q" instead.
*    It ist a problem with the Classic REXX implementation.
*
****************************************************************************/

/* load REXXUTIL funcs */
CALL RxFuncAdd "SysLoadFuncs", "RexxUtil", "SysLoadFuncs"
CALL SysLoadFuncs
NUMERIC DIGITS 12

/* init vars */
opt.sdir = ''
opt.debug = 0
opt.mout = 1
opt.fcomp = 0
opt.sync = 0
opt.nodel = 1
opt.move = 0
opt.longname = 0
opt.thres = 0
opt.tz = 0
opt.updfile = ''
opt.sayeq = 0
opt.ignoredowngrade = 0
opt.help = 0
/* autoconfig */
IF FindFilePath('XCOPY.EXE') = '' THEN
   opt.copycmd = "@COPY"
 ELSE
   opt.copycmd = "@XCOPY /H /T /R /O"
IF FindFilePath('XCOPMP.EXE') = '' THEN
   opt.compcmd = "@ECHO n|COMP"
 ELSE
   opt.compcmd = "@XCOMP"
opt.delcmd = "@DEL"
opt.movecmd = "@MOVE"

dlerr = 0
source.2 = ''
opt.xcl = 0

/****************************************************************************
   parse command line
****************************************************************************/
kstat = 0
params = ARG(1)
DO WHILE params \= ''
   IF LEFT(params,1) = '"' THEN
      PARSE VAR params '"'param'" 'params
    ELSE
      PARSE VAR params param params
   CALL ParseArg(param)
   END

IF kstat < 1 | opt.help THEN DO
   SAY "Compare and/or synchronize files and directories"
   SAY "(C) Marcel Mller 1999"
   SAY
   SAY "Usage: DSYNC filespec1 filespec2 [options]"
   SAY
   SAY "compare results:"
   SAY " *O  file in tree 1 newer"
   SAY " O*  file in tree 2 newer"
   SAY " *-  file in tree 2 not found"
   SAY " -*  file in tree 1 not found"
   SAY " >>  file in tree 1 is longer with same time stamp"
   SAY " <<  file in tree 2 is longer with same time stamp"
   SAY " <>  files have different content with same time stamp"
   SAY " EE  error during compare"
   SAY
   SAY "options:"
   SAY " /S      include subdirectories"
   SAY " /F      compare file content"
   SAY " /V      compare result in text format instead of the default as noted above"
   SAY " /Q      quiet, i.e. no output to stdout"
   SAY " /Y      synchronize files"
   SAY " /Y1     update only 1st tree"
   SAY " /Y2     update only 2nd tree"
   SAY " /Ufile  create update file/directory in filespec1 using file as index"
   SAY " /ZI     shortcut for /S /Q /U, create initial update file"
   SAY " /ZP     shortcut for /S /Q /U /Y1, pack update packet"
   SAY " /ZU     shortcut for /S /Q /U /Y2, unpack update packet"
   SAY " /ZA     shortcut for /S /Q /U /Y /M, fully automaic resync of update packet"
   SAY " /D      delete files when removed in the other tree"
   SAY " /M      move files from update directory instead of copying"
   SAY " /Tthres threshold in seconds for time compares"
   SAY " /T      use timezone corretion"
   SAY " /I      ignore downgrades (see reference)"
   SAY " /L      use PMSHELL copy routine and .LONGNAME EA"
   EXIT 48
   END

IF kstat > 2 THEN
   CALL Error 49, "Syntax error: more the 2 source paths."

/* get root path & filespec */
DO i = 1 TO 2
   /* translate '/' */
   source.i = TRANSLATE(source.i, '\', '/')
   /* get pattern */
   name.i = FILESPEC("N", source.i)
   /* check if directory */
   IF (name.i \= '') & (VERIFY(name.i, "*?", 'M') = 0) THEN DO
      CALL SysFileTree source.i, tree.i, "DO"
      IF tree.i \= 0 THEN DO
         IF opt.debug THEN
            SAY source.i" seems to be a directory, '\' added."
         source.i = source.i'\'
         name.i = ''
         END
      END
   /* get root directory */
   root.i = DirSpec(source.i)
   CALL SETLOCAL
   IF root.i = '' THEN
      root.i = DIRECTORY()
    ELSE IF LEFT(root.i, 2) \= '\\' THEN /* not for UNC path */
      root.i = DIRECTORY(root.i)
   SAY root.i
   IF RIGHT(root.i, 1) \= '\' THEN
      root.i = root.i'\'
   CALL ENDLOCAL
   offset.i = LENGTH(root.i) +23
   END

/* some defaults */
IF name.1 = '' THEN
   name.1 = '*'
IF name.2 = '' THEN
   name.2 = name.1
IF LENGTH(opt.updfile) = 1 THEN
   opt.updfile = root.1||"UPDATE.DIR"
 ELSE
   opt.updfile = SUBSTR(opt.updfile, 2)

IF opt.debug THEN DO
   DO i = 1 TO 2
      SAY "Filespec "i": root = "root.i", pattern = "name.i
      END
   SAY "Timezone: "opt.tz
   END

/* some consistency checks */
IF opt.updfile \= '' & opt.fcomp THEN
   CALL Error 34, "/U and /F are mutual exclusive."
IF opt.longname & opt.fcomp THEN
   CALL Error 34, "/L and /F are currently not supported together."
IF opt.sync = 3 & \opt.nodel THEN
   CALL Error 34, "/Y and /D are mutual exclusive."
IF root.1 = root.2 THEN
   CALL Error 32, "Cannot compare files to themselves."
IF name.1 \= name.2 THEN
   CALL Error 32, "Cannot compare directories with different file patterns."
tmp = MIN(LENGTH(root.1), LENGTH(root.2))
IF LEFT(root.1, tmp) = LEFT(root.2, tmp) THEN
   CALL Error 32, "Source path "root.1" is not independent of source path "root.2"."


/****************************************************************************
   get file lists
****************************************************************************/
DO i = 1 TO 2
   IF i = 1 & opt.updfile \= '' THEN DO
      tree.1 = 0
      IF opt.sync = 0 THEN ITERATE
      /* fetch data from update file */
      IF STREAM(opt.updfile, 'C', 'open read') \= 'READY:' THEN DO
         IF opt.sync = 3 THEN DO
            CALL LINEOUT STDERR, "Could not find index file "opt.updfile"."
            CALL LINEOUT STDERR, "In case of the first call with /ZA you may want to create an initial one."
            CALL LINEOUT STDERR, "Create initial index file [y|n] ?"
            tmp = TRANSLATE(LINEIN())
            IF LEFT(tmp,1) = 'Y' THEN DO
               opt.sync = 0
               ITERATE
               END
            END
         CALL Error 20, "Failed to read update index file "opt.updfile"."
         END
      voldesc.1 = ''
      DO j = 1
         DO FOREVER
            tmp = TRANSLATE(LINEIN(opt.updfile), '\', '/')
            IF LEFT(tmp, 1) \= ':' THEN LEAVE
            /* volume descriptor */
            voldesc.1 = SUBSTR(tmp, 2)
            END
         PARSE VAR tmp time size name
         IF name = '' THEN
            LEAVE
         tree.1.j = RIGHT(time+0, 10, 0)" "RIGHT(size+0, 10, 0)" "STRIP(name)
         END
      tree.1 = j -1
      CALL STREAM opt.updfile, 'C', 'close'
      offset.1 = 23
      END
    ELSE
      CALL ScanTree i, root.i||name.i
   /* sort file list */
   CALL fqsort i, 1, tree.i
   /*DO j = 1 TO tree.i
      SAY tree.i.j
      END*/
   END
/* volume descriptors */
IF opt.updfile \= '' THEN DO
   tmp = FILESPEC('D', root.2)
   IF tmp \= '' THEN
      voldesc.2 = '<'WORD(SysDriveInfo(tmp), 4)'> 'root.2
    ELSE
      voldesc.2 = root.2
   IF voldesc.1 \= '' & opt.sync // 3 \= 0 & COMPARE(TRANSLATE(voldesc.1), TRANSLATE(voldesc.2), '00'x) = 0 THEN DO
      CALL LINEOUT STDERR, "Update tree is possibly the same then at the last call."
      CALL LINEOUT STDERR, "You propably want to use /za or /y."
      CALL LINEOUT STDERR, "Continue anyway [y|n] ?"
      IF LEFT(TRANSLATE(LINEIN()), 1) \= 'Y' THEN
         EXIT 18
      END
   END

/****************************************************************************
   compare ...
****************************************************************************/
i1 = 0
i2 = 0
name1 = NextName(1)
name2 = NextName(2)
DO WHILE name1 \= '' | name2 \= ''
   /* filenames equal ? */
   name = TRANSLATE(name1)
   IF '_'name = '_'TRANSLATE(name2) THEN DO
      /* compare time stamp */
      diff = ABS(LEFT(tree.1.i1, 10) - LEFT(tree.2.i2, 10))
      IF diff > opt.thres THEN DO
         IF diff < 86400 & diff // 3600 = 0 THEN
            CALL DaylightErr
         IF opt.debug THEN
            SAY "Time diff: "name1" "||(LEFT(tree.1.i1, 10) - LEFT(tree.2.i2, 10))
         IF LEFT(tree.1.i1, 10) < LEFT(tree.2.i2, 10) THEN
            CALL DoItem 'O*', name2
          ELSE
            CALL DoItem '*O', name1
         END
      /* compare size */
       ELSE IF SUBSTR(tree.1.i1, 12, 10) \= SUBSTR(tree.2.i2, 12, 10) THEN DO
         IF SUBSTR(tree.1.i1, 12, 10) > SUBSTR(tree.2.i2, 12, 10) THEN
            CALL DoItem '>>', name1
          ELSE
            CALL DoItem '<<', name2
         END
      /* compare content */
       ELSE IF opt.fcomp THEN DO
         opt.compcmd' "'PCTesc(SUBSTR(tree.1.i1, 23)'" "'SUBSTR(tree.2.i2, 23))'" >NUL'
         IF rc = 1 | rc = 5 THEN
            CALL DoItem '<>', name2
          ELSE IF rc > 1 THEN
            CALL DoItem 'EE', name2, rc
         END
       ELSE
         CALL DoItem '==', name2
    next:
      name1 = NextName(1)
      name2 = NextName(2)
      END
   /* filenames not equal */
    ELSE IF "_"name < "_"TRANSLATE(name2) THEN DO
      CALL DoItem '*-', name1
      name1 = NextName(1)
      END
    ELSE DO
      CALL DoItem '-*', name2
      name2 = NextName(2)
      END
   END

/****************************************************************************
   create update file
****************************************************************************/
IF opt.updfile \= '' & opt.sync \= 2 & \opt.debug THEN DO
   IF opt.sync = 3 THEN
      /* rescan tree to include recent changes */
      CALL ScanTree 2, root.2||name.2
   IF opt.sync \= 0 THEN
      /* remove old file if any */
      opt.delcmd' "'PCTesc(opt.updfile)'" 2>NUL'
    ELSE IF opt.sync = 0 & STREAM(opt.updfile, 'c', 'query exists') \= '' THEN
      CALL Error 14, "Update file "opt.updfile" alredy exists. Maybe you forgot /Y."
   IF STREAM(opt.updfile, 'C', 'open write') \= 'READY:' THEN
      CALL Error 27, "Failed to create update index file "opt.updfile"."
   CALL LINEOUT opt.updfile, ':'voldesc.2
   DO i = 1 TO tree.2
      CALL LINEOUT opt.updfile, LEFT(tree.2.i, 10)+0" "SUBSTR(tree.2.i, 12, 10)+0" "SUBSTR(tree.2.i, offset.2)
      END
   CALL STREAM opt.updfile, 'C', 'close'
   END

EXIT 0


/****************************************************************************
   main functions
****************************************************************************/

/* parse command line argument
   ParseArg(arg)
*/
ParseArg: PROCEDURE EXPOSE opt. source. kstat
   IF ARG(1) = '' THEN
      RETURN
   IF LEFT(ARG(1), 1) = '@' THEN DO
      /* option file */
      param = SUBSTR(ARG(1), 2)
      IF STREAM(param, 'c', 'open read') \= 'READY:' THEN
         CALL Error 30, "Failed to open indirect file "param"."
      DO WHILE STREAM(param) = 'READY'
         CALL ParseArg LINEIN(param)
         END
      CALL STREAM param, 'c', 'close'
      RETURN
      END
   IF LEFT(ARG(1), 1) \= '/' & LEFT(ARG(1), 1) \= '-' THEN DO
      /* filename */
      kstat = kstat +1
      source.kstat = ARG(1)
      RETURN
      END
   /* option */
   param = TRANSLATE(SUBSTR(ARG(1), 2))
   SELECT
    WHEN param = '_' THEN
      opt.debug = 1 /* debug mode */
    WHEN param = 'S' THEN
      opt.sdir = 'S' /* subdirectories */
    WHEN LEFT(param, 1) = 'X' THEN DO
      opt.xcl = opt.xcl + 1  /* file eopt.xcludes */
      i = opt.xcl
      opt.xcl.i = TRANSLATE(SUBSTR(param, 2))
      END
    WHEN param = 'V0' | param = 'Q' THEN
      opt.mout = 0  /* quiet */
    WHEN param = 'V1' THEN
      opt.mout = 1  /* normal output */
    WHEN param = 'V2' | param = 'V' THEN
      opt.mout = 2  /* verbose output */
    WHEN param = 'F' THEN
      opt.fcomp = 1 /* compare file content */
    WHEN param = 'D' THEN
      opt.nodel = 0 /* no delete */
    WHEN param = 'M' THEN
      opt.move = 1  /* move from 2nd */
    WHEN LEFT(param, 1) = 'T' THEN
      IF LENGTH(param) = 1 THEN
         opt.tz = InitTimezone() /* timezone corrections */
       ELSE IF param = 'TD' THEN DO
         opt.tz = InitTimezone() /* timezone corrections */
         opt.tz = "0 "SUBSTR(opt.tz, WORDINDEX(opt.tz, 2))
         END
       ELSE IF param = 'TZ' THEN
         opt.tz = WORD(InitTimezone(), 1) /* timezone corrections */
       ELSE
         opt.thres = SUBSTR(param, 2) /* threshold */
    WHEN param = 'L' THEN
      opt.longname = 1 /* longname EA */
    WHEN param = 'Y1' THEN
      opt.sync = 1
    WHEN param = 'Y2' THEN
      opt.sync = 2
    WHEN param = 'Y' | param = 'Y3' THEN
      opt.sync = 3
    WHEN param = 'I' THEN
      opt.ignoredowngrade = 1
    WHEN LEFT(param, 1) = 'U' THEN
      opt.updfile = param
    WHEN LEFT(param, 2) = 'ZI' | LEFT(param, 2) = 'Z0' THEN DO
      opt.sdir = 'S' /* subdirectories */
      opt.mout = 0  /* quiet */
      opt.updfile = SUBSTR(param, 2)
      END
    WHEN LEFT(param, 2) = 'ZP' | LEFT(param, 2) = 'Z1' THEN DO
      opt.sdir = 'S' /* subdirectories */
      opt.mout = 0  /* quiet */
      opt.sync = 1
      opt.updfile = SUBSTR(param, 2)
      END
    WHEN LEFT(param, 2) = 'ZU' | LEFT(param, 2) = 'Z2' THEN DO
      opt.sdir = 'S' /* subdirectories */
      opt.mout = 0  /* quiet */
      opt.sync = 2
      opt.updfile = SUBSTR(param, 2)
      END
    WHEN LEFT(param, 2) = 'ZA' | LEFT(param, 2) = 'Z3' THEN DO
      opt.sdir = 'S' /* subdirectories */
      opt.mout = 0  /* quiet */
      opt.move = 1  /* move files from tree2 */
      opt.sync = 3
      opt.updfile = SUBSTR(param, 2)
      END
    WHEN param = '?' THEN
      opt.help = 1
    OTHERWISE
      CALL Error 44, "Invalid option "ARG(1)"."
      END
   RETURN

/* scan source tree
   ScanTree(tree#, filespec)
*/
ScanTree: PROCEDURE EXPOSE tree. opt. monsum
   i = ARG(1)
   /* get file tree */
   DROP tree.i
   CALL SysFileTree ARG(2), tree.i, "FT"opt.sdir
   tree.i = tree.i.0
   IF tree.i = 0 THEN
      CALL Error 8, "Source tree "i" does not exist or is empty."
   /* convert internal format */
   DO j = 1 TO tree.i
      PARSE VAR tree.i.j time size attrib file
      file = STRIP(file)
      PARSE VALUE STREAM(file, 'c', 'query datetime') WITH month'-'day'-'year' 'hour':'min':'sec
      IF opt.longname THEN
         IF SysGetEA(file, '.LONGNAME', 'tmp') = 0 THEN
            IF tmp \= '' THEN
               file = DirSpec(file)'\'SUBSTR(tmp, 5)
      /* IF opt.debug THEN
         SAY "F: "tree.i.j */
      /* Y2k fix */
      tree.i.j = RIGHT(ToUnixTime(RIGHT(year+30,2)+1970, month, day, STRIP(hour), min, sec), 10, 0)" "RIGHT(size, 10, 0)" "file
      END
   RETURN

/* handle difference
   DoItem(operation, filename [, additional info])
   operation:
     '*O' file in tree 1 newer
     'O*' file in tree 2 newer
     '*-' file in tree 2 not found
     '-*' file in tree 1 not found
     '>>' file in tree 1 is longer with same time stamp
     '<<' file in tree 2 is longer with same time stamp
     '<>' files have different content with same time stamp
     'EE' error during compare
     '==' files are identical (as far as checked)
   filename:
     filename including relative path to root
   additional info:
     in case of an error (operation = 'EE') this is written to the screen
*/
DoItem: PROCEDURE EXPOSE opt. root.
   IF opt.mout = 1 THEN DO
      /* machine readable output */
      IF (opt.sayeq | ARG(1) \= '==') & \(opt.ignoredowngrade & ((opt.sync = 1 & ARG(1) = '*O') | (opt.sync = 2 & ARG(1) = 'O*'))) THEN
         SAY ARG(1)" "ARG(2)
      END
    ELSE IF opt.mout = 2 THEN
      /* normal output */
      SELECT
       WHEN ARG(1) = 'O*' THEN
         SAY "File "ARG(2)" in "root.1" is older than in "root.2"."
       WHEN ARG(1) = '*O' THEN
         SAY "File "ARG(2)" in "root.1" is newer than in "root.2"."
       WHEN ARG(1) = '-*' THEN
         SAY "File "ARG(2)" is not found in "root.1"."
       WHEN ARG(1) = '*-' THEN
         SAY "File "ARG(2)" is not found in "root.2"."
       WHEN ARG(1) = '>>' THEN
         SAY "File "ARG(2)" in "root.1" is bigger."
       WHEN ARG(1) = '<<' THEN
         SAY "File "ARG(2)" in "root.1" is smaller."
       WHEN ARG(1) = '<>' THEN
         SAY "File "ARG(2)" has different content."
       WHEN ARG(1) = 'EE' THEN
         SAY "Error "ARG(3)" during compare of "ARG(2)"."
       WHEN ARG(1) = '==' THEN
         IF opt.sayeq THEN
            SAY "File "ARG(2)" is identical in both trees."
       OTHERWISE
         SAY ARG(1)" "ARG(2)
         END

   IF opt.sync = 0 THEN
      RETURN
   /* synchronize */
   SELECT
    WHEN RIGHT(ARG(1), 1) = '*' THEN
      /* file in tree 1 is older or does not exist */
      IF opt.sync \= 2 THEN
         CALL Warning CopyFile(ARG(2), root.2, root.1)
       ELSE IF ARG(1) = 'O*' THEN DO
         IF \opt.ignoredowngrade THEN
            CALL Warning "Cannot handle downgrade of file "ARG(2)" automatically."
         END
       ELSE IF ARG(1) = '-*' & \opt.nodel THEN
         CALL Warning DeleteFile(ARG(2), root.2)
    WHEN LEFT(ARG(1), 1) = '*' THEN
      /* file in tree 2 is older or does not exist */
      IF opt.sync \= 1 THEN DO
         IF \opt.move | opt.updfile = '' THEN
            CALL Warning CopyFile(ARG(2), root.1, root.2)
          ELSE
            CALL Warning MoveFile(ARG(2), root.1, root.2)
         END
       ELSE IF ARG(1) = '*O' THEN DO
         IF \opt.ignoredowngrade THEN
            CALL Warning "Cannot handle downgrade of file "ARG(2)" automatically."
         END
       ELSE IF ARG(1) = '*-' & \opt.nodel & opt.updfile = '' THEN
         CALL Warning DeleteFile(ARG(2), root.1)
    WHEN ARG(1) = '==' THEN
      IF opt.sync \= 1 & \opt.nodel & opt.updfile \= '' & opt.move THEN
         CALL DeleteFile ARG(2), root.1
    OTHERWISE
      CALL Warning "Synchronize failed ("ARG(1)"): "ARG(2)
      END
   RETURN

/* fetch next name from list index
   NextName(index)
*/
NextName: PROCEDURE EXPOSE i1 i2 tree. opt. offset.
   i = ARG(1)
   DO p = VALUE('i'i) +1 TO tree.i
      CALL VALUE 'i'i, p
      name = SUBSTR(tree.i.p, offset.i)
      /* check exclude list */
      IF \CheckXcl(TRANSLATE(name)) THEN
         RETURN name
      END
   RETURN ''

/* check if name is in exclude list
   CheckXcl(name)
*/
CheckXcl: PROCEDURE EXPOSE opt.
   DO i = 1 TO opt.xcl
      IF StringMatchQ(ARG(1), opt.xcl.i) THEN
         RETURN 1
      END
   RETURN 0


/****************************************************************************
   file functions
****************************************************************************/

/* remove file
   DeleteFile(file, source)
   removes file source||file and remove the directory if empty
*/
DeleteFile: PROCEDURE EXPOSE opt. root.
   file = ARG(2)ARG(1)
   IF opt.debug THEN DO
      SAY "DEL: "file
      RETURN ''
      END
   IF opt.longname THEN DO
      IF SysDestroyObject(file) = 0 THEN
         RETURN "Removal of "file" failed."
      END
    ELSE DO
      opt.delcmd' "'PCTesc(file)'" 2>NUL'
      IF RC \= 0 THEN
         RETURN "Removal of "file" failed (rc = "RC")."
      END
   /* remove directory if empty */
   CALL CheckDirRM file
   RETURN ''

/* copy file
   CopyFile(file, source, destination)
   copies source||file to destination||file
   If the destination directory does not exist it is created.
*/
CopyFile: PROCEDURE EXPOSE opt.
   dst = DirSpec(ARG(3)ARG(1))
   CALL Warning CheckDir(dst)
   IF opt.debug THEN DO
      SAY "COPY: "ARG(2)ARG(1)" "dst
      RETURN ''
      END
   IF opt.longname THEN DO
      IF SysCopyObject(ARG(2)ARG(1), dst) = 0 THEN
         RETURN "Copy of "ARG(2)ARG(1)" to "dst" failed."
      END
    ELSE DO
      opt.copycmd' "'PCTesc(ARG(2)ARG(1)'" "'dst)'" >NUL'
      IF RC \= 0 THEN
         RETURN "Copy of "ARG(2)ARG(1)" to "dst" failed (rc = "RC")."
      END
   RETURN ''

/* move file
   MoveFile(file, source, destination)
   moves source||file to destination||file
   If the destination directory does not exist it is created.
*/
MoveFile: PROCEDURE EXPOSE opt. root.
   dst = DirSpec(ARG(3)ARG(1))
   CALL Warning CheckDir(dst)
   IF opt.debug THEN DO
      SAY "MOVE: "ARG(2)ARG(1)" "dst
      RETURN ''
      END
   IF opt.longname THEN DO
      IF SysMoveObject(ARG(2)ARG(1), dst) = 0 THEN
         RETURN "Move of "ARG(2)ARG(1)" to "dst" failed."
      END
/*  I had no luck with move
    ELSE IF LEFT(ARG(2), POS(':',ARG(2)':')) = LEFT(ARG(3), POS(':',ARG(3))) THEN DO
      opt.movecmd' "'PCTesc(ARG(2)ARG(1)'" "'dst)'" >NUL'
      IF RC \= 0 THEN
         RETURN "Move of "ARG(2)ARG(1)" to "dst" failed (rc = "RC")."
      END */
    ELSE DO
      /* well, somtimes move may not work ... */
      opt.copycmd' "'PCTesc(ARG(2)ARG(1)'" "'dst)'" >NUL'
      IF RC \= 0 THEN
         RETURN "Copy (move) of "ARG(2)ARG(1)" to "dst" failed (rc = "RC")."
      opt.delcmd' "'PCTesc(ARG(2)ARG(1))'" >NUL'
      IF RC \= 0 THEN
         CALL Warning "Removal of "ARG(2)ARG(1)" during move to "dst" failed (rc = "RC")."
      END
   /* remove directory if empty */
   CALL CheckDirRM ARG(2)ARG(1)
   RETURN ''

/* create directory if neccessary
   CheckDir(path)
   This function calls itself recursively.
*/
CheckDir: PROCEDURE EXPOSE opt.
   IF ARG(1) = '' THEN RETURN ''
   IF VALUE('opt.dir.'TRANSLATE(ARG(1),,'   !"#$%&'||"'()+,-./:;=@[\]^`{|}~",'_'), '') = '' THEN
      RETURN ''
   CALL SysFileTree ARG(1), res, 'DO'
   IF res.0 \= 0 THEN
      RETURN ''
   /* directory does not exist */
   CALL CheckDir DirSpec(ARG(1))
   IF RESULT \= '' THEN
      RETURN RESULT
   IF opt.debug THEN
      SAY "MKDIR: "ARG(1)
    ELSE DO
      CALL SysMkDir ARG(1)
      IF RESULT \= 0 THEN
         RETURN "Create directory "ARG(1)" failed ("RESULT")"
      END
   RETURN ''

/* remove parent directory if empty
   CheckDirRM(path)
*/
CheckDirRM: PROCEDURE EXPOSE root.
   dir = DirSpec(ARG(1))
   IF dir = '' | VERIFY(RIGHT(dir, 1), ': ') = 0 | dir'\' = root.1 | dir'\' = root.2 THEN
      RETURN
   DO WHILE SysRmDir(dir) = 0
      dir = DirSpec(dir)
      END
   RETURN


/****************************************************************************
   helper functions
****************************************************************************/

/* sort directory tree
   fqsort(index, first, last)
   index tree table index
     1 = tree.1..., 2 = tree.2...
   first, last
     sort range tree.index.first to tree.index.last
   This function calls itself recursively.
*/
fqsort: PROCEDURE EXPOSE tree. offset.
   list = ARG(1)
   lo = ARG(2)
   hi = ARG(3)
   mi = (hi + lo) %2
   m = TRANSLATE(SUBSTR(tree.list.mi, offset.list))
   /*SAY lo hi mi m*/
   DO WHILE hi >= lo
      DO WHILE TRANSLATE(SUBSTR(tree.list.lo, offset.list)) < m
         lo = lo +1
         END
      DO WHILE TRANSLATE(SUBSTR(tree.list.hi, offset.list)) > m
         hi = hi -1
         END
      IF hi >= lo THEN DO
         tmp = tree.list.lo
         tree.list.lo = tree.list.hi
         tree.list.hi = tmp
         lo = lo +1
         hi = hi -1
         END
      END
   IF hi > ARG(2) THEN
      CALL fqsort list, ARG(2), hi
   IF ARG(3) > lo THEN
      CALL fqsort list, lo, ARG(3)
   RETURN

/* wildcard match
   StringMatchQ(string, template)
*/
StringMatchQ: PROCEDURE
   str = ARG(1)
   tmp = ARG(2)
 redo:
   IF LENGTH(str) = 0 THEN
      RETURN VERIFY(tmp, '*') = 0 /* '' matches only if template consists only of '*' */
   p = VERIFY(tmp, '?*', 'M')
   IF p = 0 THEN
      RETURN COMPARE(str, tmp, '*') = 0 /* no wildcards => compare strings */
   IF COMPARE(LEFT(str, p-1), LEFT(tmp, p-1, '*')) \= 0 THEN
      RETURN 0 /* compare of non-wildcard section failed */
   j = p
   DO i = p TO VERIFY(tmp' ', '*?',, p +1) -1 /* count # of ? */
      IF SUBSTR(tmp, i, 1) = '?' THEN
         j = j +1
      END
   IF j > LENGTH(str)+1 THEN
      RETURN 0 /* more ? than length of str */
   tmp = SUBSTR(tmp, i)
   str = SUBSTR(str, j)
   IF i = j THEN
      SIGNAL redo
   /* '*' */
   IF LENGTH(tmp) = 0 THEN
      RETURN 1 /* nothing after '*' */
   DO p = 1
      p = POS(LEFT(tmp, 1), str, p)
      IF p = 0 THEN
         RETURN 0 /* character after '*' not found */
      IF StringMatchQ(SUBSTR(str, p+1), SUBSTR(tmp, 2)) THEN
         RETURN 1 /* got it! */
      END

InitTimezone: PROCEDURE
   monsum = "0 31 59 90 120 151 181 212 243 273 304 334 365"
   PARSE VALUE VALUE('TZ',,'OS2ENVIRONMENT') WITH tz','sm','sw','sd','st','em','ew','ed','et','offset
   tz = SUBSTR(tz, 4)
   p = VERIFY(tz, '-0123456789:')
   IF p = 0 THEN
      offset = 0
    ELSE
      tz = LEFT(tz, p -1)
   PARSE VAR tz hour':'min':'sec
   IF min = '' THEN min = 0
   IF sec = '' THEN sec = 0
   IF hour = '' THEN hour = 0
   IF hour < 0 THEN DO
      min = -min
      sec = -sec
      END
   tz = 60*(60*hour + min) + sec
   IF offset = 0 THEN
      RETURN tz
   IF sm = '' THEN sm = 4
   IF sw = '' THEN sw = 1
   IF sd = '' THEN sd = 0
   IF st = '' THEN st = 3600
   IF em = '' THEN em = 10
   IF ew = '' THEN ew = -1
   IF ed = '' THEN ed = 0
   IF et = '' THEN et = 7200
   IF offset = '' THEN offset = 3600
   IF sw = 0 THEN
      st = st + 86400 * (WORD(monsum, sm) + sd-1) + offset
    ELSE DO
      IF sw < 0 THEN
         sm = sm +1
       ELSE
         sw = sw -1
      st = st + 86400 * (WORD(monsum, sm) + 7*sw) + offset
      sd = 11 - sd
      END
   IF ew = 0 THEN
      et = et + 86400 * (WORD(monsum, em) + ed-1) + offset
    ELSE DO
      IF ew < 0 THEN
         em = em +1
       ELSE
         ew = ew -1
      et = et + 86400 * (WORD(monsum, em) + 7*ew) + offset
      ed = 11 - ed
      END
   RETURN tz offset st st + 86400 * (sm >= 3) sd et et + 86400 * (em >= 3) ed

/* convert to unix time
   ToUnixTime(year, month, day, hour, minute, second)
*/
ToUnixtime: PROCEDURE EXPOSE opt.
   monsum = "0 31 59 90 120 151 181 212 243 273 304 334 365"
   year70 = ARG(1) - 1970;
   /* check
   if ( tm->Sec > 59U || tm->Min > 59U || tm->Hour > 23U
       || (unsigned)year70 >= 136U || (tm->Month-1U) > 11U
       || (tm->Day-1) >= monlen[tm->Month-1]
       || (tm->Day == 29 && tm->Month == 2 && (year70&3)) )
      return false;*/
   ysec = ARG(6) + 60*ARG(5) + 3600*ARG(4) + 86400*(ARG(3)-1 + WORD(monsum, ARG(2)) + (ARG(2) >= 3 & year70//4 == 2))
   ydays = 365*year70 + (year70+1)%4
   PARSE VAR opt.tz delta offset sn ss sd en es ed
   tmp = ysec + 86400*ydays + delta
   IF offset = '' THEN RETURN tmp
   /* daylight saving correction */
   IF year70//4 == 2 THEN DO
      sn = ss
      en = es
      END
   IF sd \= 0 THEN
      sn = sn + 86400*(6 - (ydays + sn%86400 + sd) // 7)
   IF ed \= 0 THEN
      en = en + 86400*(6 - (ydays + en%86400 + ed) // 7)
   IF BITXOR(ysec < sn, ysec >= en) \= 1 THEN /* XOR -> even works on the southern hemisphere */
      tmp = tmp - offset
/*   SAY uc-tmp uc tmp ARG(4)':'ARG(5)':'ARG(6) ARG(3)'.'ARG(2)'.'ARG(1)*/
   RETURN tmp

DaylightErr: PROCEDURE EXPOSE dlerr
   dlerr = dlerr + 1
   IF dlerr \= 3 THEN RETURN
   CALL LINEOUT STDERR, "Warning: several time stamps differ exactly by a few hole hours."
   CALL LINEOUT STDERR, " Maybe there is a problem with the time zone or with daylight saving."
   RETURN

/* quote % characters */
PCTesc: PROCEDURE
   tmp = ARG(1)
   p = POS('%', tmp)
   DO WHILE p \= 0
      tmp = INSERT('%', tmp, p)
      p = POS('%', tmp, p+2)
      END
   RETURN tmp

/* locate file (in PATH) */
FindFilePath: PROCEDURE
   /* first check .\ */
   tmp = STREAM(ARG(1), 'c', 'query exists')
   IF tmp = '' THEN
      tmp = SysSearchPath('PATH', ARG(1))
   RETURN tmp

DirSpec: PROCEDURE
   RETURN STRIP(FILESPEC('D', ARG(1))FILESPEC('P', ARG(1)), 'T', '\')

Error: PROCEDURE
   CALL LINEOUT STDERR, ARG(2)
   EXIT ARG(1)

Warning: PROCEDURE
   IF ARG(1) \= '' THEN
      CALL LINEOUT STDERR, ARG(1)
   RETURN ARG(1)

